/*
 * regex: Regular expression pattern match library
 * 
 * by A.ITO, December 1989
 */

#ifdef DEBUG
#include <sys/types.h>
#include <malloc.h>
#endif
#include <ctype.h>

#ifdef NOBCOPY
void bcopy(void*,void*,int);
void bzero(void*,int);
#else
#include <strings.h>
#endif

#define REGEX_MAX	64
#define STORAGE_MAX	256

#ifndef NULL
#define NULL	0
#endif

#define	RE_NORMAL	0
#define RE_MATCHMODE	0x07
#define RE_ANY		0x01
#define RE_WHICH	0x02
#define RE_EXCEPT	0x04
#define RE_ANYTIME	0x08
#define RE_BEGIN	0x10
#define RE_END		0x20
#define RE_IGNCASE      0x40
#define RE_ENDMARK	0x80

#define KANJI1(c)	((c)&0x80)	/* for EUC */

typedef unsigned short longchar;

typedef struct {
  longchar       *pattern;
  char            mode;
}               regexchar;

#define RE_WHICH_RANGE	0xffff

static regexchar CompiledRegex[REGEX_MAX];
static longchar Cstorage[STORAGE_MAX];
static longchar *st_ptr;
static char    *position, *lposition;

static          regmatch(regexchar *, char *, int, char **);
static          regmatch1(regexchar *, longchar);
static          matchWhich(longchar *, longchar);


/*
 * regexCompile: compile regular expression
 */
char           *
regexCompile(char *ex,int igncase)
{
  char           *p;
  longchar       *r;
  regexchar      *re = CompiledRegex - 1;
  int             m;

  st_ptr = Cstorage;
  for (p = ex; *p != '\0'; p++) {
    switch (*p) {
    case '.':
      re++;
      re->pattern = NULL;
      re->mode = RE_ANY;
      break;
    case '$':
      re++;
      re->pattern = NULL;
      re->mode = RE_END;
      break;
    case '^':
      re++;
      re->pattern = NULL;
      re->mode = RE_BEGIN;
      break;
    case '*':
      re->mode |= RE_ANYTIME;
      break;
    case '[':
      r = st_ptr;
      if (*++p == '^') {
	p++;
	m = RE_EXCEPT;
      } else
	m = RE_WHICH;
      while (*p != ']') {
	if (*p == '\\') {
	  *(st_ptr++) = *(p + 1);
	  p += 2;
	} else if (*p == '-') {
	  *(st_ptr++) = RE_WHICH_RANGE;
	  p++;
	} else if (*p == '\0')
	  return "Missing ]";
	else if (KANJI1(*p)) {
	  bcopy((void*)p, (void*)st_ptr, sizeof(longchar));
	  st_ptr++;
	  p += sizeof(longchar);
	} else
	  *(st_ptr++) = *(p++);
      }
      *(st_ptr++) = '\0';
      re++;
      re->pattern = r;
      re->mode = m;
      break;
    case '\\':
      p++;
    default:
      re++;
      if (KANJI1(*p)) {
	bcopy((void*)p, (void*)st_ptr, sizeof(longchar));
	p += sizeof(longchar) - 1;
      } else
	*st_ptr = *p;
      re->pattern = st_ptr;
      st_ptr++;
      re->mode = RE_NORMAL;
      if (igncase)
        re->mode |= RE_IGNCASE;
    }
    if (st_ptr >= &Cstorage[STORAGE_MAX] ||
	re >= &CompiledRegex[REGEX_MAX])
      return "Regular expression too long";
  }
  re++;
  re->mode = RE_ENDMARK;
  return NULL;
}

/*
 * regexMatch: match regular expression
 */
int
regexMatch(char *str, int firstp)
{
  char           *p;

  if (str == NULL)
    return 0;
  position = NULL;
  for (p = str; *p != '\0'; p++) {
    switch (regmatch(CompiledRegex, p, firstp && (p == str), &lposition)) {
    case 1:
      position = p;
      return 1;
    case -1:
      position = NULL;
      return -1;
    }
  }
  return 0;
}

/*
 * matchedPosition: last matched position
 */
void
matchedPosition(char **first, char **last)
{
  *first = position;
  *last = lposition;
}

/*
 * Intermal routines
 */
static int
regmatch(regexchar * re, char *str, int firstp, char **lastpos)
{
  char           *p = str;
  char           *lpos, *llpos = NULL;
  longchar        k;

#ifdef DEBUG
  debugre(re, str);
#endif
  while ((re->mode & RE_ENDMARK) == 0) {
    if (re->mode & RE_BEGIN) {
      if (!firstp)
	return 0;
      re++;
    } else if (re->mode & RE_ANYTIME) {
      short           matched = 0, ok = 0;
      do {
	if (regmatch(re + 1, p, firstp, &lpos) == 1) {
	  llpos = lpos;
	  matched = 1;
	} else if (matched) {
	  ok = 1;
	  break;
	}
	if (*p == '\0') {
	  if (matched)
	    ok = 1;
	  break;
	}
	if (KANJI1(*p)) {
	  bcopy((void*)p, (void*)&k, sizeof(longchar));
	  if (regmatch1(re, k)) {
	    if (lastpos != NULL)
	      *lastpos = llpos;
	    p += sizeof(longchar);
	  } else if (matched)
	    ok = 1;
	  else
	    break;
	} else {
	  if (regmatch1(re, *p)) {
	    p++;
	    if (lastpos != NULL)
	      *lastpos = llpos;
	  } else if (matched)
	    ok = 1;
	  else
	    break;
	}
      } while (!ok);
      if (lastpos != NULL)
	*lastpos = llpos;
      return ok;
    } else if (re->mode & RE_END) {
      if (lastpos != NULL)
	*lastpos = p;
      return (*p == '\0');
    } else {
      int             a;
      if (KANJI1(*p)) {
	bcopy((void*)p, (void*)&k, sizeof(longchar));
	a = regmatch1(re, k);
	p += sizeof(longchar);
      } else
	a = regmatch1(re, *(p++));
      if (!a)
	return 0;
      else
	re++;
    }
  }
  if (lastpos != NULL)
    *lastpos = p;
  return 1;
}

static int
regmatch1(regexchar * re, longchar c)
{
  switch (re->mode & RE_MATCHMODE) {
    case RE_ANY:
#ifdef DEBUG
    printf("%c vs any. -> 1\n", c);
#endif
    return 1;
  case RE_NORMAL:
#ifdef DEBUG
    printf("RE=%c vs %c -> %d\n", *re->pattern, c, *re->pattern == c);
#endif
    if (re->mode & RE_IGNCASE) {
      if (*re->pattern < 127 && c < 127 &&
          isalpha(*re->pattern) && isalpha(c))
        return tolower(*re->pattern)==tolower(c);
      else
        return *re->pattern == c;
    }
    else
      return (*re->pattern == c);
  case RE_WHICH:
    return matchWhich(re->pattern, c);
  case RE_EXCEPT:
    return !matchWhich(re->pattern, c);
  }
  return 0;
}

static int
matchWhich(longchar * pattern, longchar c)
{
  longchar       *p = pattern;
  int             ans = 0;

#ifdef DEBUG
  printf("RE pattern = %s char=%c", pattern, c);
#endif
  while (*p != '\0') {
    if (*(p + 1) == RE_WHICH_RANGE && *(p + 2) != '\0') {	/* Char class. */
      if (*p <= c && c <= *(p + 2)) {
	ans = 1;
	break;
      }
      p += 3;
    } else {
      if (*p == c) {
	ans = 1;
	break;
      }
      p++;
    }
  }
#ifdef DEBUG
  printf(" -> %d\n", ans);
#endif
  return ans;
}

#ifdef DEBUG
char           *
lc2c(longchar * x)
{
  static char     y[100];
  int             i = 0;

  while (x[i]) {
    if (x[i] == RE_WHICH_RANGE)
      y[i] = '-';
    else
      y[i] = x[i];
    i++;
  }
  y[i] = '\0';
  return y;
}

void
debugre(re, s)
  regexchar      *re;
  char           *s;
{
  for (; !(re->mode & RE_ENDMARK); re++) {
    if (re->mode & RE_BEGIN) {
      printf("Begin ");
      continue;
    } else if (re->mode & RE_END) {
      printf("End ");
      continue;
    }
    if (re->mode & RE_ANYTIME)
      printf("Anytime-");

    switch (re->mode & RE_MATCHMODE) {
    case RE_ANY:
      printf("Any ");
      break;
    case RE_NORMAL:
      printf("Match-to'%c' ", *re->pattern);
      break;
    case RE_WHICH:
      printf("One-of\"%s\" ", lc2c(re->pattern));
      break;
    case RE_EXCEPT:
      printf("Other-than\"%s\" ", lc2c(re->pattern));
      break;
    default:
      printf("Unknown ");
    }
  }
  putchar('\n');
}

#endif
